"""
__author__ = Hagai Hargil
"""
import pathlib
from pysight.main_multiscaler_readout import main_data_readout
from pysight.tkinter_gui_multiscaler import GUIApp
from pysight.tkinter_gui_multiscaler import verify_gui_input

config_filename = r''

path = pathlib.Path(config_filename).parent
all_lst_files = path.glob('*.lst')

gui = GUIApp()
gui.root.mainloop()
verify_gui_input(gui)
df_list = []
movie_list = []
outputs_list = []
censored_list = []

for idx, lst_file in enumerate(all_lst_files):
    df, movie, outputs, censored = main_data_readout(gui)
    df_list.append(df)
    movie_list.append(movie)
    outputs_list.append(outputs)
    censored_list.append(censored)
    label = 6
    data, labels = censored.learn_histograms(label)
    filename = r'/data/Lior/Multiscaler data/06 June 2017/TrainedWeights/17p_label_{}.npy'.format(label)
    import numpy as np

    with open(filename, 'wb') as f:
        np.save(f, data)





"""
__author__ = Hagai Hargil
"""
from unittest import TestCase
import pandas as pd


class TestFrame(TestCase):
    """
    Tests for the Frame class
    """
    list_of_file_names = [r'..\TAG ON start channel, galvo on stop 2 - gain 480.lst',
                          r'..\fixed sample - XY image - 500 gain - 1 second acquisition.lst'
                          r'..\live mouse  100 um deep with 62p TAG010.lst'
                          r'..\multiscaler_check_code_2_channels_0-1_sec.lst']

    data = pd.read_hdf('df_for_test.h5')

    def test_first_line_time(self):
        self.fail()

    def test_last_line_time(self):
        self.fail()

    def test_first_event_time(self):
        self.fail()

    def test_max_delta_of_lines(self):
        self.fail()

    def test_create_hist_edges(self):
        self.fail()

    def test_create_hist(self):
        self.fail()

    def test_display(self):
        self.fail()



import unittest
from os import sep


class TestLstTools(unittest.TestCase):
    from pysight.lst_tools import Analysis


    analysis = Analysis()
    analysis.run()
    list_of_file_names = ['tests_data' + sep + 'data_for_tests.lst']
    list_of_real_start_loc = [1749]
    list_of_real_time_patch = ['32']
    list_of_real_range = [80000000 * 2 ** 4]

    def test_complete_workflow(self):

        from pysight.fileIO_tools import read_lst
        from pysight.lst_tools import tabulate_input
        from pysight.lst_tools import determine_data_channels
        from pysight.lst_tools import allocate_photons  # TODO: Split this test
        from pysight.timepatch_switch import ChoiceManager
        import numpy as np

        list_of_first_event = ['0100000060d9']
        list_of_first_times = [1549]

        list_of_returned_events = []
        list_of_returned_times = []

        for idx, fname in enumerate(self.list_of_file_names):
            arr = read_lst(fname, self.list_of_real_start_loc[idx], timepatch=self.list_of_real_time_patch[idx])
            first_event = arr[0]
            dict_of_slices = ChoiceManager().process(self.list_of_real_time_patch[idx])
            list_of_returned_events.append(first_event)
            dict_of_inputs = {'PMT1': '001', 'Frames': '110', 'Lines': '010'}
            df_sorted = tabulate_input(arr, dict_of_slices, self.list_of_real_range[idx], dict_of_inputs)
            list_of_returned_times.append(df_sorted['abs_time'][0])

        self.assertEqual(list_of_first_event[0], list_of_returned_events[0][:-2])
        self.assertEqual(list_of_first_times, list_of_returned_times)

        dict_of_data = determine_data_channels(df=df_sorted, dict_of_inputs=dict_of_inputs,
                                               num_of_frames=2, x_pixels=512, y_pixels=512)
        df_allocated = allocate_photons(dict_of_data=dict_of_data)
        calculated_data = set(df_allocated.iloc[0].values)
        data_to_check_allocation = {59, 678024}
        index_data = set(df_allocated.index.values[0])
        real_data = {2557, 677965, 0}

        self.assertTrue(np.array_equal(data_to_check_allocation, calculated_data))
        self.assertEqual(real_data, index_data)

    # def test_create_frame_array_normal_input(self):  # TODO: Fix this test
    #     import numpy as np
    #     from pysight.lst_tools import create_frame_array
    #
    #     last_event = int(1e3)
    #     num_of_events = 10
    #     check_linspace = list(np.linspace(0, last_event, num=num_of_events, endpoint=False))
    #
    #     self.assertEqual(check_linspace, list(create_frame_array(last_event_time=last_event,
    #                                                              num_of_frames=num_of_events)))

    def test_create_frame_array_negative_input(self):
        from pysight.lst_tools import create_frame_array

        last_event = int(-1e4)
        num_of_frames = 10

        with self.assertRaises(ValueError):
            arr = create_frame_array(last_event, num_of_frames)

    def test_create_line_array_normal_input(self):
        import numpy as np
        from pysight.lst_tools import create_line_array

        last_event = int(1e4)
        num_of_events = 10
        num_of_frames = 4

        check_linspace = list(np.linspace(0, last_event, num_of_events * num_of_frames))

        self.assertEqual(check_linspace, list(create_line_array(last_event, num_of_events, num_of_frames)))

    def test_create_line_array_negative_input(self):
        from pysight.lst_tools import create_line_array

        last_event = int(-1e4)
        num_of_events = 10
        num_of_frames = 4

        with self.assertRaises(ValueError):
            arr = create_line_array(last_event, num_of_events, num_of_frames)

    def test_determine_data_channels_empty_df(self):
        import pandas as pd
        from pysight.lst_tools import determine_data_channels

        test_df = pd.DataFrame(data=[])
        empty_dict = {}

        with self.assertRaises(ValueError):
            test = determine_data_channels(test_df, empty_dict)

    def test_determine_data_channels_short_dict(self):
        from pysight.lst_tools import determine_data_channels
        import pandas as pd

        dict1 = {'a': 1, 'b': 2}
        df = pd.DataFrame([1, 2])

        with self.assertRaises(KeyError):
            test = determine_data_channels(df, dict1)


    def test_create_inputs_dict_empty_gui(self):
        from pysight.fileIO_tools import create_inputs_dict

        trial_gui = None

        with self.assertRaises(ValueError):
            dict1 = create_inputs_dict(trial_gui)


if __name__ == '__main__':
    unittest.main()



import pandas as pd
import numpy as np
import attr
from typing import Dict, Tuple
from numba import jit, int64, uint64
import warnings
from pysight.apply_df_funcs import get_lost_bit_np, get_lost_bit_tag, iter_string_hex_to_bin, convert_hex_to_int, convert_hex_to_bin
from pysight.validation_tools import validate_line_input, validate_frame_input, \
    validate_laser_input, validate_created_data_channels, rectify_photons_in_uneven_lines
from attr.validators import instance_of



@attr.s(slots=True)
class Analysis(object):
    """
    Create pipeline of analysis of lst files.
    """
    # TODO: Input validations
    # TODO: Variable documentation
    timepatch          = attr.ib(validator=instance_of(str))
    data_range         = attr.ib(validator=instance_of(int))
    dict_of_inputs     = attr.ib()
    data               = attr.ib()
    is_binary          = attr.ib(validator=instance_of(bool))
    num_of_frames      = attr.ib(validator=instance_of(int))
    x_pixels           = attr.ib(validator=instance_of(int))
    y_pixels           = attr.ib(validator=instance_of(int))
    laser_freq         = attr.ib(validator=instance_of(float))
    binwidth           = attr.ib(validator=instance_of(float))
    dict_of_slices_hex = attr.ib()
    dict_of_slices_bin = attr.ib()
    bidir              = attr.ib(validator=instance_of(int))
    tag_freq           = attr.ib(validator=instance_of(float))
    tag_pulses         = attr.ib(validator=instance_of(int))
    phase              = attr.ib(validator=instance_of(float))
    use_tag_bits       = attr.ib(validator=instance_of(int))
    laser_offset       = attr.ib(validator=instance_of(float))
    keep_unidir        = attr.ib(default=False)
    df_allocated       = attr.ib(init=False)
    dict_of_data       = attr.ib(init=False)

    def run(self):
        """ Pipeline of analysis """
        if self.is_binary:
            df_after_timepatch = self.tabulate_input_binary()
        else:
            df_after_timepatch = self.tabulate_input_hex()

        print('Sorted dataframe created. Starting setting the proper data channel distribution...')
        self.dict_of_data, line_delta = self.determine_data_channels(df=df_after_timepatch)
        print('Channels of events found. Allocating photons to their frames and lines...')
        self.df_allocated = self.allocate_photons(dict_of_data=self.dict_of_data, line_delta=line_delta)
        print('Relative times calculated. Creating Movie object...')
        # Censor correction addition:
        if 'Laser' not in self.dict_of_data.keys():
            self.dict_of_data['Laser'] = 0

    @staticmethod
    def hex_to_bin_dict():
        """
        Create a simple dictionary that maps a hex input into a 4 letter binary output.
        :return: dict
        """
        diction = \
            {
                '0': '0000',
                '1': '0001',
                '2': '0010',
                '3': '0011',
                '4': '0100',
                '5': '0101',
                '6': '0110',
                '7': '0111',
                '8': '1000',
                '9': '1001',
                'a': '1010',
                'b': '1011',
                'c': '1100',
                'd': '1101',
                'e': '1110',
                'f': '1111',
            }
        return diction

    @property
    def offset(self):
        return int(np.floor(self.laser_offset * 10**-9 / self.binwidth))

    def determine_data_channels(self, df: pd.DataFrame=None) -> Tuple:
        """ Create a dictionary that contains the data in its ordered form."""

        if df.empty:
            raise ValueError('Received dataframe was empty.')

        dict_of_data = {}
        for key in self.dict_of_inputs:
            relevant_values = df.loc[df['channel'] == self.dict_of_inputs[key], 'abs_time']
            # NUMBA SORT NOT WORKING:
            # sorted_vals = numba_sorted(relevant_values.values)
            # dict_of_data[key] = pd.DataFrame(sorted_vals, columns=['abs_time'])
            if 'PMT1' == key or 'PMT2' == key:
                dict_of_data[key] = relevant_values.reset_index(drop=True)
            else:
                dict_of_data[key] = relevant_values.sort_values().reset_index(drop=True)

        # Validations
        dict_of_data, line_delta = validate_line_input(dict_of_data=dict_of_data, num_of_lines=self.y_pixels,
                                                       num_of_frames=self.num_of_frames)
        dict_of_data = validate_frame_input(dict_of_data=dict_of_data, line_delta=line_delta,
                                            num_of_lines=self.y_pixels, binwidth=self.binwidth)
        try:
            dict_of_data['Laser'] = validate_laser_input(dict_of_data['Laser'], laser_freq=self.laser_freq,
                                                         binwidth=self.binwidth, offset=self.offset)
        except KeyError:
            pass

        validate_created_data_channels(dict_of_data)
        return dict_of_data, line_delta

    def allocate_photons(self, dict_of_data=None, line_delta: float = -1) -> pd.DataFrame:
        """
        Returns a dataframe in which each photon is a part of a frame, line and possibly laser pulse
        :param dict_of_data: All events data, distributed to its input channel
        :param gui: Input GUI
        :return: pandas.DataFrame
        """
        from pysight.tag_tools import interpolate_tag

        # Preparations
        irrelevant_keys = {'PMT1', 'PMT2', 'TAG Lens'}
        relevant_keys = set(dict_of_data.keys()) - irrelevant_keys

        df_photons = dict_of_data['PMT1']  # TODO: Support more than one data channel
        df_photons = pd.DataFrame(df_photons, dtype='uint64')  # before this change it's a series with a name, not column head
        column_heads = {'Lines': 'time_rel_line_pre_drop', 'Frames': 'time_rel_frames', 'Laser': 'time_rel_pulse'}

        # Unidirectional scan - create fake lines
        if not self.bidir:
            dict_of_data = self.add_unidirectional_lines(dict_of_data=dict_of_data, line_delta=line_delta)

        # Main loop - Sort lines and frames for all photons and calculate relative time
        for key in reversed(sorted(relevant_keys)):
            sorted_indices = numba_search_sorted(dict_of_data[key].values, df_photons['abs_time'].values)
            try:
                df_photons[key] = dict_of_data[key].loc[sorted_indices].values
            except KeyError:
                warnings.warn('All computed sorted_indices were "-1" for key {}. Trying to resume...'.format(key))

            df_photons.dropna(how='any', inplace=True)
            df_photons[key] = df_photons[key].astype(np.uint64)
            df_photons[column_heads[key]] = df_photons['abs_time'] - df_photons[key] # relative time of each photon in
            #                                                                          accordance to the line\frame\laser pulse
            if 'Lines' == key:
                df_photons = rectify_photons_in_uneven_lines(df=df_photons,
                                                             sorted_indices=sorted_indices[sorted_indices >= 0],
                                                             lines=dict_of_data['Lines'], bidir=self.bidir,
                                                             phase=self.phase, keep_unidir=self.keep_unidir)

            if 'Laser' != key:
                df_photons[key] = df_photons[key].astype('category')
            df_photons.set_index(keys=key, inplace=True, append=True, drop=True)

        # Closure
        assert np.all(df_photons.values >= 0)  # finds NaNs as well
        # Deal with TAG lens interpolation
        try:
            tag = dict_of_data['TAG Lens']
        except KeyError:
            pass
        else:
            print('Interpolating TAG lens data...')
            df_photons = interpolate_tag(df_photons=df_photons, tag_data=tag, tag_freq=self.tag_freq,
                                         binwidth=self.binwidth, tag_pulses=self.tag_pulses)
            print('TAG lens interpolation finished.')

        # df_photons.drop(['abs_time'], axis=1, inplace=True)

        return df_photons

    def process_chan_edge(self, struct_of_data):
        """
        Simple processing scheme for the channel and edge data.
        """
        bin_array = np.array(iter_string_hex_to_bin("".join(struct_of_data.data)))
        edge      = self.slice_string_arrays(bin_array, start=0, end=1)
        channel   = self.slice_string_arrays(bin_array, start=1, end=4)

        return edge, channel

    def tabulate_input_hex(self) -> pd.DataFrame:
        """
        Reformat the read hex data into a dataframe.
        """

        for key in list(self.dict_of_slices_hex.keys())[1:]:
            self.dict_of_slices_hex[key].data = self.slice_string_arrays(self.data, self.dict_of_slices_hex[key].start,
                                                                         self.dict_of_slices_hex[key].end)

        if not self.use_tag_bits:
            self.dict_of_slices_hex.pop('tag')

        # Channel and edge information
        edge, channel = self.process_chan_edge(self.dict_of_slices_hex.pop('chan_edge'))
        # TODO: Timepatch == '3' is not supported because of this loop.

        if self.dict_of_slices_hex['lost'] is True:
            for key in list(self.dict_of_slices_hex.keys())[1:]:
                if self.dict_of_slices_hex[key].needs_bits:
                    list_with_lost = iter_string_hex_to_bin("".join(self.dict_of_slices_hex[key].data))
                    step_size = self.dict_of_slices_hex[key].end - self.dict_of_slices_hex[key].start
                    if key == 'tag':
                        list_of_losts, self.dict_of_slices_hex[key].processed = get_lost_bit_tag(list_with_lost,
                                                                                                 step_size,
                                                                                                 len(self.data))
                    else:
                        list_of_losts, self.dict_of_slices_hex[key].processed = get_lost_bit_np(list_with_lost,
                                                                                                step_size,
                                                                                                len(self.data))
                else:
                    if key == 'tag':
                        self.dict_of_slices_hex[key].processed = convert_hex_to_bin(self.dict_of_slices_hex[key].data)
                    else:
                        self.dict_of_slices_hex[key].processed = convert_hex_to_int(self.dict_of_slices_hex[key].data)
        else:
            for key in list(self.dict_of_slices_hex.keys())[1:]:
                if key == 'tag':
                    self.dict_of_slices_hex[key].processed = convert_hex_to_bin(self.dict_of_slices_hex[key].data)
                else:
                    self.dict_of_slices_hex[key].processed = convert_hex_to_int(self.dict_of_slices_hex[key].data)

        # Reformat data
        df = pd.DataFrame(channel, columns=['channel'], dtype='category')
        df['edge'] = edge
        df['edge'] = df['edge'].astype('category')

        try:
            df['tag'] = self.dict_of_slices_hex['tag'].processed
        except KeyError:
            pass

        try:
            df['lost'] = list_of_losts  # TODO: Currently the LOST bit is meaningless
            df['lost'] = df['lost'].astype('category')
        except NameError:
            pass

        df['abs_time'] = np.uint64(0)

        if 'sweep' in self.dict_of_slices_hex:
            df['abs_time'] = self.dict_of_slices_hex['abs_time'].processed + (
                (self.dict_of_slices_hex['sweep'].processed - 1) * self.data_range)
        else:
            df['abs_time'] = self.dict_of_slices_hex['abs_time'].processed

        # Before sorting all photons make sure that no input is missing from the user. If it's missing
        # the code will ignore this channel, but not raise an exception
        actual_data_channels = set(df['channel'].cat.categories.values)
        if actual_data_channels != set(self.dict_of_inputs.values()):
            warnings.warn("Channels that were inserted in GUI don't match actual data channels recorded. \n"
                          "The list files contains data in the following channels: {}.".format(actual_data_channels))

        assert np.all(df['abs_time'].values >= 0)

        return df

    @staticmethod
    def slice_string_arrays(arr: np.array, start: int, end: int) -> np.array:
        """
        Slice an array of strings efficiently.
        Based on http://stackoverflow.com/questions/39042214/how-can-i-slice-each-element-of-a-numpy-array-of-strings
        with modifications for Python 3.
        """
        b = arr.view('U1').reshape(len(arr), -1)[:, start:end]
        return np.fromstring(b.tostring(), dtype='U' + str(end - start))

    def tabulate_input_binary(self) -> pd.DataFrame:
        """
        Reformat the read binary data into a dataframe.
        """
        num_of_lines = self.data.shape[0]

        for key in self.dict_of_slices_bin:
            cur_data = self.data[:, self.dict_of_slices_bin[key].start:self.dict_of_slices_bin[key].end]
            try:
                zero_arr = np.zeros(num_of_lines, self.dict_of_slices_bin[key].cols)

            except AttributeError:  # No cols field since number of bits is a multiple of 8
                # self.dict_of_slices_bin[key].data_as_
                pass

    @staticmethod
    def add_unidirectional_lines(dict_of_data: Dict, line_delta: float = -1):
        """
        For unidirectional scans fake line signals have to be inserted.
        :param dict_of_data: All data
        :param line_delta: difference between lines
        :return:
        """

        if line_delta == -1:
            raise ValueError('Line delta variable was miscalculated.')

        length_of_lines       = dict_of_data['Lines'].shape[0]
        new_line_arr          = np.zeros(length_of_lines * 2 - 1)
        new_line_arr[::2]     = dict_of_data['Lines'].values
        new_line_arr[1::2]    = dict_of_data['Lines'].rolling(window=2).mean()[1:]

        dict_of_data['Lines'] = pd.Series(new_line_arr, name='Lines', dtype='uint64')
        return dict_of_data


@jit(nopython=True, cache=True)
def numba_sorted(arr: np.array) -> np.array:
    """
    Sort an array with Numba. CURRENTLY NOT WORKING
    """

    arr.sort()
    return arr.astype(np.uint64)


@jit((int64[:](uint64[:], uint64[:])), nopython=True, cache=True)
def numba_search_sorted(input_sorted, input_values):
    """ Numba-powered searchsorted function. """
    return np.searchsorted(input_sorted, input_values) - 1



"""
__author__ = Hagai Hargil
"""
import attr
import numpy as np
import pandas as pd
from attr.validators import instance_of
from pysight.movie_tools import Volume, Movie
from collections import deque, namedtuple
from typing import Tuple, Union
from numba import jit, uint64, uint8, int64


@attr.s(slots=True)
class CensorCorrection(object):
    movie            = attr.ib(validator=instance_of(Movie))
    reprate          = attr.ib(validator=instance_of(float))
    binwidth         = attr.ib(validator=instance_of(float))
    laser_offset     = attr.ib(validator=instance_of(float))
    all_laser_pulses = attr.ib()

    @property
    def bins_bet_pulses(self) -> int:
        return int(np.ceil(1 / (self.reprate * self.binwidth)))

    @property
    def offset(self):
        return int(np.floor(self.laser_offset * 10**-9 / self.binwidth))

    def __gen_laser_pulses_deque(self) -> np.ndarray:
        """
        If data has laser pulses - return them. Else - simulate them with an offset
        """
        start_time = 0
        step = self.bins_bet_pulses
        volumes_in_movie = self.movie.gen_of_volumes()

        if self.all_laser_pulses == 0:  # no 'Laser' data was recorded
            for vol in volumes_in_movie:
                yield np.arange(start=start_time+self.offset, stop=vol.end_time,
                                step=step, dtype=np.uint64)

        else:
            for vol in volumes_in_movie:
                yield self.all_laser_pulses[(self.all_laser_pulses >= vol.abs_start_time-step) &
                                            (self.all_laser_pulses <= vol.end_time+step)] + self.offset

    def __get_bincount_deque(self):
        print("Movie object created. Generating the bincount deque...")
        bincount_deque = deque()
        laser_pulses_deque = self.__gen_laser_pulses_deque()
        volumes_in_movie = self.movie.gen_of_volumes()
        for idx, vol in enumerate(volumes_in_movie):
            censored = CensoredVolume(df=vol.data, vol=vol, offset=self.offset,
                                      laser_pulses=next(laser_pulses_deque))
            dig, bincount = censored.gen_bincount()
            pos_idx = np.where(dig >= 0)[0]
            dig = dig[pos_idx]
            pos_photons = censored.df.iloc[pos_idx, -1].values.T
            if len(pos_photons) == 0:
                data_dict = {'photon_hist'    : np.zeros((self.bins_bet_pulses, 1), dtype=np.uint8),
                             'bincount'       : bincount,
                             'num_empty_hists': bincount[0]}
                return data_dict

            photon_hist = np.zeros((self.bins_bet_pulses, pos_photons.shape[0]), dtype=np.uint8)
            for laser_idx, photon in enumerate(np.nditer(pos_photons)):
                start_time = censored.laser_pulses[dig[laser_idx]]
                try:
                    end_time = censored.laser_pulses[dig[laser_idx] + 1]
                except IndexError:  # photons out of laser pulses
                    continue
                else:
                    photon_hist[:, laser_idx] = np.histogram(photon, bins=np.arange(start_time, end_time + 1,
                                                                                    dtype='uint64'))[0]

            data_dict = {'photon_hist'    : photon_hist,
                         'bincount'       : bincount,
                         'num_empty_hists': bincount[0]}
            assert data_dict['num_empty_hists'] >= 0, 'Sum of bincount: {}, number of photons: {}'\
                .format(sum(bincount), laser_idx)
            bincount_deque.append(data_dict)

        return bincount_deque

    def find_temporal_structure_deque(self):
        temp_struct_deque = deque()
        laser_pulses_deque = self.__gen_laser_pulses_deque()
        volumes_in_movie = self.movie.gen_of_volumes()
        for idx, vol in enumerate(volumes_in_movie):
            censored = CensoredVolume(df=vol.data, vol=vol, offset=self.offset,
                                      laser_pulses=next(laser_pulses_deque),
                                      binwidth=self.binwidth, reprate=self.reprate)
            temp_struct_deque.append(censored.find_temp_structure())
        return temp_struct_deque

    def create_array_of_hists_deque(self):
        """
        Go through each volume in the deque and find the laser pulses for each pixel, creating a summed histogram per pixel.
        :return:
        """
        temp_struct_deque = deque()
        laser_pulses_deque = self.__gen_laser_pulses_deque()
        volumes_in_movie = self.movie.gen_of_volumes()
        for idx, vol in enumerate(volumes_in_movie):
            censored = CensoredVolume(df=vol.data, vol=vol, offset=self.offset,
                                      laser_pulses=next(laser_pulses_deque),
                                      binwidth=self.binwidth, reprate=self.reprate)
            temp_struct_deque.append(censored.gen_array_of_hists())
        return temp_struct_deque

    def __gen_labels(self, size: int, label: Union[int, float]) -> np.ndarray:
        """
        Create labels for the ML algorithm. Label value must be an integer.
        :size: Number of elements
        :return: np.ndarray
        """
        if isinstance(label, int):  # fixed power during the session
            return np.ones(size, dtype=np.uint8) * label
        elif isinstance(label, float):  # `label` contains the frequency of the triangular wave
            pass

    def learn_histograms(self, label: Union[int, float], power: int, folder_to_save: str):
        """
        Implement the machine learning algorithm on the data.
        :param label: Label of ML algorithm.
        :param power: How much power was injected to the Qubig. For saving the file.
        :return: data, labels
        """
        from sklearn import svm, metrics
        import pathlib

        # Start by generating the data and arranging it properly for the machine
        bincount = self.__get_bincount_deque()
        print("Bincount done. Adding all data to a single matrix.")

        data = np.empty((self.bins_bet_pulses, 0))
        for vol in bincount:
            data = np.concatenate((data, vol['photon_hist']), axis=1)  # the histograms with photons in them
            data = np.concatenate((data, np.zeros((self.bins_bet_pulses, vol['num_empty_hists']),
                                                  dtype=np.uint8)), axis=1)  # empty hists
        data = data.T
        n_samples = data.shape[0]
        labels = self.__gen_labels(n_samples, label)
        classifier = svm.SVC(gamma=0.001)
        labels[1] = 10  # toying around

        print("Fitting the data...")
        classifier.fit(data[:n_samples // 2], labels[:n_samples // 2])

        # Predictions
        expected = labels[n_samples // 2:]
        predicted = classifier.predict(data[n_samples // 2:])
        print("Number of samples is %s." % n_samples)
        print("Classification report for classifier %s:\n%s\n"
              % (classifier, metrics.classification_report(expected, predicted)))
        print("Confusion matrix:\n%s" % metrics.confusion_matrix(expected, predicted))

        # Save the data for future use
        folder_as_path = pathlib.Path(folder_to_save)
        filename = str(folder_as_path / "{}p_label_{}.npy".format(power, label))
        self.__save_data(data=data, filename=filename)
        return data, labels

    def __save_data(self, data: np.ndarray, filename: str):
        """
        Save the data array for future training.
        :param data: Data to be learnt.
        :param filename: Including dir
        :return:
        """
        print("Saving to {}...".format(filename))
        with open(filename, 'wb') as f:
            np.save(f, data)


@attr.s(slots=True)
class CensoredVolume(object):
    df = attr.ib(validator=instance_of(pd.DataFrame))
    vol = attr.ib(validator=instance_of(Volume))
    laser_pulses = attr.ib(validator=instance_of(np.ndarray))
    offset = attr.ib(validator=instance_of(int))
    binwidth = attr.ib(default=800e-12)
    reprate = attr.ib(default=80e6)

    def gen_bincount(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Bin the photons into their relative laser pulses, and count how many photons arrived due to each pulse.
        """
        hist, _ = np.histogram(self.vol.data.loc[:, 'time_rel_frames'].values, bins=self.laser_pulses)
        dig = np.digitize(self.vol.data.loc[:, 'time_rel_frames'].values,
                          bins=self.laser_pulses) - 1
        return dig, np.bincount(hist)

    def find_temp_structure(self) -> np.ndarray:
        """
        Generate a summed histogram of the temporal structure of detected photons.
        """
        bins = np.arange(start=0, stop=np.ceil(1/(self.reprate * self.binwidth)) + 1, step=1)
        subtracted_times, _, _ = self.sort_photons_in_pulses()
        hist, _ = np.histogram(subtracted_times, bins=bins)
        return hist

    def sort_photons_in_pulses(self):
        """
        Helper function to generate a searchsorted output of photons in laser pulses.
        """
        pulses = self.laser_pulses
        sorted_indices: np.ndarray = np.searchsorted(pulses, self.vol.data['time_rel_frames'].values) - 1
        array_of_laser_starts = pulses[sorted_indices]
        subtracted_times = self.vol.data['time_rel_frames'].values - array_of_laser_starts
        return subtracted_times, array_of_laser_starts, sorted_indices

    def gen_array_of_hists(self) -> np.ndarray:
        """
        For a specific frame, sort photons and laser pulses inside the pixels to gain
        statistics on the distribution of the photons inside the pixels.
        :return: np.ndarray of the same size as the original image. Each pixels contains
        a histogram inside it.
        """
        BinData = namedtuple('BinData', ('hist', 'pulses', 'photons'))
        all_pulses = 0
        all_photons = 0

        hist, edges = self.vol.create_hist()
        # Create a relative timestamp to the line signal for each laser pulse
        sorted_pulses = np.searchsorted(edges[0][:-1], self.laser_pulses) - 1
        pulses = pd.DataFrame(data=self.laser_pulses[np.where(sorted_pulses >= 0)[0]], columns=['time_rel_frames'])
        pulses = pulses.assign(Lines=edges[0][:-1][sorted_pulses[np.where(sorted_pulses >= 0)[0]]])
        pulses.dropna(how='any', inplace=True)
        pulses.loc[:, 'Lines'] = pulses.loc[:, 'Lines'].astype('uint64')
        pulses.loc[:, 'time_rel_line'] = pulses.loc[:, 'time_rel_frames'] - pulses.loc[:, 'Lines']
        pulses.loc[:, 'Lines'] = pulses.loc[:, 'Lines'].astype('category')
        pulses.set_index(keys=['Lines'], inplace=True, append=True, drop=True)

        # Allocate laser pulses and photons to their bins
        pulses.loc[:, 'bins_x'] = (np.digitize(pulses.loc[:, 'time_rel_frames'].values,
                                   bins=edges[0]) - 1).astype('uint16', copy=False)
        pulses.loc[:, 'bins_y'] = (np.digitize(pulses.loc[:, 'time_rel_line'].values,
                                   bins=edges[1]) - 1).astype('uint16', copy=False)
        self.vol.data.loc[:, 'bins_x'] = (np.digitize(self.vol.data.loc[:, 'time_rel_frames'].values,
                                          bins=edges[0]) - 1).astype('uint16', copy=False)
        self.vol.data.loc[:, 'bins_y'] = (np.digitize(self.vol.data.loc[:, 'time_rel_line'].values,
                                          bins=edges[1]) - 1).astype('uint16', copy=False)
        pulses.set_index(keys=['bins_x', 'bins_y'], inplace=True, append=True, drop=True)
        self.vol.data.set_index(keys=['bins_x', 'bins_y'], inplace=True, append=True, drop=True)

        # Go through each bin and histogram the photons there
        image_bincount = np.zeros_like(hist, dtype=object)
        for row in range(self.vol.x_pixels):
            row_pulses = pulses.xs(key=row, level='bins_x', drop_level=False)
            assert len(row_pulses) > 0, 'Row {} contains no pulses'.format(row)
            try:
                row_photons = self.vol.data.xs(key=row, level='bins_x', drop_level=False)
            except KeyError:  # no photons in row
                for col in range(self.vol.y_pixels):
                    final_pulses = row_pulses.xs(key=col, level='bins_y', drop_level=False)
                    hist = (np.histogram(np.array([]), bins=final_pulses.loc[:, 'time_rel_line'].values)[0])\
                        .astype('uint8')
                    cur_bincount = np.bincount(hist).astype('uint64', copy=False)
                    tot_pulses = np.sum(cur_bincount)
                    tot_photons = np.average(cur_bincount, weights=range(len(cur_bincount)))
                    image_bincount[row, col] = BinData(hist=hist, pulses=tot_pulses, photons=tot_photons)
                    all_photons += tot_photons
                    all_pulses += tot_pulses
            else:
                for col in range(self.vol.y_pixels):
                    final_pulses = row_pulses.xs(key=col, level='bins_y', drop_level=False)
                    assert len(final_pulses) > 0, 'Column {} in row {} contains no pulses'.format(col, row)
                    try:
                        final_photons = row_photons.xs(key=col, level='bins_y', drop_level=False)
                    except KeyError:  # no photons in col
                        hist = (np.histogram(np.array([]), bins=final_pulses.loc[:, 'time_rel_line'].values)[0])\
                            .astype('uint8', copy=False)
                    else:
                        # hist = (np.histogram(final_photons.loc[:, 'time_rel_line'].values,
                        #                     bins=final_pulses.loc[:, 'time_rel_line'].values)[0])\
                        #     .astype('uint8', copy=False)
                        hist = numba_histogram(final_photons.loc[:, 'time_rel_line'].values,
                                               bins=final_pulses.loc[:, 'time_rel_line'].values)\
                            .astype('uint8', copy=False)
                    finally:
                        if not np.all(hist >= 0):
                            print('WHAT IS GOING ON')
                        assert np.all(hist >= 0), 'In row {}, column {}, the histogram turned out to be negative.'.format(row, col)
                        cur_bincount = numba_bincount(hist).astype('uint64', copy=False)
                        tot_pulses = np.sum(cur_bincount)
                        tot_photons = np.average(cur_bincount, weights=range(len(cur_bincount)))
                        image_bincount[row, col] = BinData(hist=hist, pulses=tot_pulses, photons=tot_photons)
                        all_photons += tot_photons
                        all_pulses += tot_pulses

        return image_bincount, all_pulses, all_photons


@jit((int64[:](uint64[:], uint64[:])), nopython=True, cache=True)
def numba_histogram(arr: np.array, bins) -> np.array:
    return np.histogram(arr, bins)[0]

@jit((int64[:](uint8[:])), nopython=True, cache=True)
def numba_bincount(arr: np.array) -> np.array:
    return np.bincount(arr)
#
# @jit((int64[:](uint64[:], uint64[:])), nopython=True, cache=True)
# def numba_digitize(arr: np.array) -> np.array:
#     pass



"""
__author__ = Hagai Hargil
"""
from typing import List
import numpy as np



def generate_output_list(movie, gui) -> List:
    """
    Main function that generates outputs based on given parameters.
    :return List containing outputs.
    """

    output_list: List = []

    if 1 == gui.summed.get():
        output_list.append(gen_single(movie=movie, gui=gui))

    if 1 == gui.full.get():
        output_list.append(gen_array(movie=movie, gui=gui))

    if 1 == gui.tif.get():
        output_list.append(gen_tiff(movie=movie, gui=gui))

    return output_list


def gen_tiff(movie, gui) -> int:
    """
    Algorithm's output will be a tiff file.
    """
    movie.create_tif()
    print('Tiff stack created with name {}.tif'.format(gui.filename.get()[:-4]))
    return 1


def gen_array(movie, gui):
    """
    Algorithm's output will be the full array of data
    """
    data_array = movie.create_array()
    print('Data array created.')
    return data_array


def gen_single(movie, gui) -> np.ndarray:
    """
    Algorithm's output will be a single stack.
    """
    single_volume: np.ndarray = movie.create_single_volume('all')
    print(r'A summed array of all volumes\frames was created.')
    return single_volume




"""
__author__ = Hagai Hargil
"""
from typing import Dict
import pandas as pd
import numpy as np
import warnings


def validate_line_input(dict_of_data: Dict, num_of_lines: int = -1, num_of_frames: int = -1, binwidth: float = 800e-12):
    """ Verify that the .lst input of lines exists and looks fine. Create one if there's no such input. """
    if num_of_lines == -1:
        raise ValueError('No number of lines input received.')

    if num_of_frames == -1:
        raise ValueError('No number of frames received.')

    last_event_time = dict_of_data['PMT1'].max()  # TODO: Assuming only data from PMT1 is relevant here
    if 'Lines' in dict_of_data.keys():
        # Verify that the input is not corrupt
        median_of_lines = dict_of_data['Lines'].diff().abs().median()
        mean_of_lines = dict_of_data['Lines'].diff().abs().mean()
        diff_between_mean_median = abs((mean_of_lines - median_of_lines)/median_of_lines)
        if diff_between_mean_median > 0.15:  # Large diff suggesting corrupt data
            warnings.warn('The difference between the mean and median values of the line channel is {}%.\n'
                          .format(diff_between_mean_median * 100))
            line_delta = float(input('If you know the expected time (in seconds) between subsequent lines please write it. Else, write 0:\n'))
            line_delta *= binwidth
            if line_delta <= 0:
                raise ValueError('Line data was corrupt.')
            # Create a new line array
            line_array = create_line_array(last_event_time=last_event_time, num_of_lines=num_of_lines,
                                           num_of_frames=num_of_frames)
            dict_of_data['Lines'] = pd.Series(line_array, name='abs_time', dtype='uint64')
        else:
            # Data is valid. Check whether we need a 0-time line event
            line_delta = median_of_lines
            zeroth_line_delta = np.abs(dict_of_data['Lines'][0] - line_delta)/line_delta
            if zeroth_line_delta < 0.05:
                dict_of_data['Lines'] = pd.Series([0], name='Lines', dtype='uint64') \
                    .append(dict_of_data['Lines'], ignore_index=True)

    else:  # create our own line array
        line_array = create_line_array(last_event_time=last_event_time, num_of_lines=num_of_lines,
                                       num_of_frames=num_of_frames)
        dict_of_data['Lines'] = pd.Series(line_array, name='abs_time', dtype='uint64')
        line_delta = last_event_time/(num_of_lines * int(num_of_frames))

    return dict_of_data, line_delta


def validate_frame_input(dict_of_data: Dict, binwidth, line_delta: int = -1, num_of_lines: int = -1):
    if line_delta == -1:
        raise ValueError('No line delta input received.')

    if num_of_lines == -1:
        raise ValueError('No number of lines received.')

    if 'Frames' in dict_of_data.keys():
        dict_of_data['Frames'] = pd.Series([0], name='abs_time', dtype='uint64')\
            .append(dict_of_data['Frames'], ignore_index=True)
    else:
        last_event_time = int(dict_of_data['Lines'].max() + line_delta)
        frame_array = create_frame_array(lines=dict_of_data['Lines'], last_event_time=last_event_time,
                                         pixels=num_of_lines)
        dict_of_data['Frames'] = pd.Series(frame_array, name='abs_time', dtype='uint64')

    return dict_of_data


def create_frame_array(lines: pd.Series=None, last_event_time: int=None,
                       pixels: int=None) -> np.ndarray:
    """Create a pandas Series of start-of-frame times"""

    if last_event_time is None or pixels is None or lines.empty:
        raise ValueError('Wrong input detected.')

    if last_event_time <= 0:
        raise ValueError('Last event time is zero or negative.')

    num_of_recorded_lines = lines.shape[0]
    actual_num_of_frames = max(num_of_recorded_lines // pixels, 1)

    if num_of_recorded_lines < pixels:
        array_of_frames = np.linspace(start=0, stop=last_event_time, num=int(actual_num_of_frames), endpoint=False)
    else:
        unnecess_lines = num_of_recorded_lines % pixels
        array_of_frames = lines.iloc[0 : int(num_of_recorded_lines-unnecess_lines) : pixels]


    return np.array(array_of_frames)


def create_line_array(last_event_time: int=None, num_of_lines=None, num_of_frames=None) -> np.ndarray:
    """Create a pandas Series of start-of-line times"""

    if (last_event_time is None) or (num_of_lines is None) or (num_of_frames is None):
        raise ValueError('Wrong input detected.')

    if (num_of_lines <= 0) or (num_of_frames <= 0):
        raise ValueError('Number of lines and frames has to be positive.')

    if last_event_time <= 0:
        raise ValueError('Last event time is zero or negative.')

    total_lines = num_of_lines * int(num_of_frames)
    line_array = np.arange(start=0, stop=last_event_time, step=last_event_time/total_lines)
    return line_array


def validate_created_data_channels(dict_of_data: Dict):
    """
    Make sure that the dictionary that contains all data channels makes sense.
    """
    assert {'PMT1', 'Lines', 'Frames'} <= set(dict_of_data.keys())  # A is subset of B

    if dict_of_data['Frames'].shape[0] > dict_of_data['Lines'].shape[0]:  # more frames than lines
        raise UserWarning('More frames than lines, consider replacing the two.')

    try:
        if dict_of_data['TAG Lens'].shape[0] < dict_of_data['Lines'].shape[0]:
            raise UserWarning('More lines than TAG pulses, consider replacing the two.')
    except KeyError:
        pass

    try:
        if dict_of_data['Laser'].shape[0] < dict_of_data['Lines'].shape[0] or \
           dict_of_data['Laser'].shape[0] < dict_of_data['Frames'].shape[0]:
            raise UserWarning('Laser pulses channel contained less ticks than the Lines or Frames channel.')
    except KeyError:
        pass

    try:
        if dict_of_data['Laser'].shape[0] < dict_of_data['TAG Lens'].shape[0]:
            raise UserWarning('Laser pulses channel contained less ticks than the TAG lens channel.')
    except KeyError:
        pass


def validate_laser_input(pulses, laser_freq: float, binwidth: float, offset: int) -> pd.Series:
    """
    Create an orderly laser pulse train.
    :param pulses:
    :param laser_freq:
    :return:
    """
    import warnings

    diffs = pulses.diff()
    pulses_final = pulses[(diffs <= np.ceil((1 / (laser_freq * binwidth)))) &
                          (diffs >= np.floor((1 / (laser_freq * binwidth))))]\
        .reset_index(drop=True) + offset
    if len(pulses_final) < 0.9 * len(pulses):
        warnings.warn("More than 10% of pulses were filtered due to bad timings. Make sure the laser input is fine.")

    pulses_final[0] = pulses[1]
    return pulses_final


def rectify_photons_in_uneven_lines(df: pd.DataFrame, sorted_indices: np.array, lines: pd.Series, bidir: bool = True,
                                    phase: float = 0, keep_unidir: bool = False):
    """
    "Deal" with photons in uneven lines. Unidir - currently throws them away.
    Bidir = flips them over.
    """
    uneven_lines = np.remainder(sorted_indices, 2)
    if bidir:
        time_rel_line = pd.Series(range(df.shape[0]), dtype='int64', name='time_rel_line')
        time_rel_line.loc[uneven_lines == 0] = df.loc[uneven_lines == 0, 'time_rel_line_pre_drop'].values
        # Reverse the relative time of the photons belonging to the uneven lines,
        # by subtracting their relative time from the start time of the next line
        lines_to_subtract_from = lines.loc[sorted_indices[uneven_lines.astype(bool)] + 1].values
        events_to_subtract = df.loc[np.logical_and(uneven_lines, 1), 'abs_time'].values
        time_rel_line.iloc[uneven_lines.nonzero()[0]] = lines_to_subtract_from - events_to_subtract \
            + (np.sin(phase) * lines[1])  # introduce phase delay between lines
        df.insert(loc=len(df.columns), value=time_rel_line.values, column='time_rel_line')

    if not bidir and not keep_unidir:
        df = df.drop(df.index[uneven_lines == 1])
        df['time_rel_line'] = df['time_rel_line_pre_drop']

    if not bidir and keep_unidir:  # Unify the excess rows and photons in them into the previous row
        sorted_indices[np.logical_and(uneven_lines, 1)] -= 1
        df['Lines'] = lines.loc[sorted_indices].values

    df.drop(['time_rel_line_pre_drop'], axis=1, inplace=True)
    df = df[df.loc[:, 'time_rel_line'] >= 0]

    return df



"""
Created on Thu Oct 13 09:37:02 2016

__author__: Hagai Hargil
"""


def main_data_readout(gui):
    """
    Main function that reads the lst file and processes its data.
    """
    from pysight.fileIO_tools import FileIO
    from pysight.lst_tools import Analysis
    from pysight.movie_tools import Movie
    from pysight import timepatch_switch
    from pysight.censor_tools import CensorCorrection
    from pysight.output_tools import generate_output_list

    # Read the file
    cur_file = FileIO(filename=gui.filename.get(), debug=gui.debug.get(), input_start=gui.input_start.get(),
                      input_stop1=gui.input_stop1.get(), input_stop2=gui.input_stop2.get())
    cur_file.run()

    # Create input structures
    dict_of_slices_hex = timepatch_switch.ChoiceManagerHex().process(cur_file.timepatch)
    # dict_of_slices_bin = timepatch_switch.ChoiceManagerBinary().process(cur_file.timepatch)  # Not supported

    # Process events into dataframe
    analyzed_struct = Analysis(timepatch=cur_file.timepatch, data_range=cur_file.data_range,
                               dict_of_inputs=cur_file.dict_of_input_channels, data=cur_file.data,
                               is_binary=cur_file.is_binary, num_of_frames=gui.num_of_frames.get(),
                               x_pixels=int(gui.x_pixels.get()), y_pixels=int(gui.y_pixels.get()),
                               laser_freq=float(gui.reprate.get()), binwidth=float(gui.binwidth.get()),
                               dict_of_slices_hex=dict_of_slices_hex, dict_of_slices_bin=None,
                               bidir=gui.bidir.get(), tag_freq=float(gui.tag_freq.get()),
                               tag_pulses=int(gui.tag_pulses.get()), phase=gui.phase.get(),
                               keep_unidir=gui.keep_unidir.get(), use_tag_bits=gui.tag_bits.get(),
                               laser_offset=gui.offset.get())
    analyzed_struct.run()

    # Create a movie object
    final_movie = Movie(data=analyzed_struct.df_allocated, x_pixels=int(gui.x_pixels.get()),
                        y_pixels=int(gui.y_pixels.get()), z_pixels=int(gui.z_pixels.get()),
                        reprate=float(gui.reprate.get()), name=gui.filename.get(),
                        binwidth=float(gui.binwidth.get()), bidir=gui.bidir.get(),
                        fill_frac=gui.fill_frac.get())

    # Censor correction part
    censored = CensorCorrection(movie=final_movie, reprate=gui.reprate.get(),
                                binwidth=gui.binwidth.get(), laser_offset=gui.offset.get(),
                                all_laser_pulses=analyzed_struct.dict_of_data['Laser'])

    # Find out what the user wanted and output it
    print('======================================================= \nOutputs:\n--------')
    output_list = generate_output_list(final_movie, gui)

    return analyzed_struct.df_allocated, final_movie, output_list, censored


def run():
    """
    Run the entire script.
    """
    from pysight.tkinter_gui_multiscaler import GUIApp
    from pysight.tkinter_gui_multiscaler import verify_gui_input

    gui = GUIApp()
    gui.root.mainloop()
    verify_gui_input(gui)
    return main_data_readout(gui)


if __name__ == '__main__':
    df, movie, outputs, censored  = run()



"""
__author__ = Hagai Hargil
"""
import pandas as pd
from typing import Dict, List
import numpy as np
import attr


@attr.s(slots=True)
class FileIO(object):
    """
    Manage pipeline of file IO process
    """
    filename    = attr.ib(validator=attr.validators.instance_of(str))
    debug       = attr.ib(validator=attr.validators.instance_of(int))
    input_start = attr.ib(validator=attr.validators.instance_of(str))
    input_stop1 = attr.ib(validator=attr.validators.instance_of(str))
    input_stop2 = attr.ib(validator=attr.validators.instance_of(str))
    is_binary   = attr.ib(init=False)
    timepatch   = attr.ib(init=False)
    data_range  = attr.ib(init=False)
    start_of_data_pos = attr.ib(init=False)
    dict_of_input_channels = attr.ib(init=False)
    list_of_recorded_data_channels = attr.ib(init=False)
    data = attr.ib(init=False)

    def run(self):
        # Open file and find the needed parameters
        self.is_binary = self.determine_binary()
        self.timepatch = self.get_timepatch()
        if self.timepatch == '3' and not self.is_binary:
            raise NotImplementedError(
                'Timepatch value "3" is currently not supported for hex files. Please message the package owner.')
        self.data_range = self.get_range()

        self.start_of_data_pos = self.get_start_pos()
        self.dict_of_input_channels = self.create_inputs_dict()
        self.list_of_recorded_data_channels = self.find_active_channels()
        self.compare_recorded_and_input_channels()
        num_of_items = self.determine_num_of_items()
        self.data = self.read_lst(num_of_items=num_of_items)
        print('File read. Sorting the file according to timepatch...')

    def determine_num_of_items(self):
        if self.debug == 0:
            num_of_items = -1
            read_string = 'Reading file "{}"...'.format(self.filename)
        else:
            num_of_items = 0.5e6
            read_string = '[DEBUG] Reading file "{}"...'.format(self.filename)

        print(read_string)
        return num_of_items

    @staticmethod
    def create_data_length_dict():
        """
        CURRENTLY DEPRECATED
        :return:
        """
        dict_of_data_length = {
                "0": 16,
                "5": 32,
                "1": 32,
                "1a": 48,
                "2a": 48,
                "22": 48,
                "32": 48,
                "2": 48,
                "5b": 64,
                "Db": 64,
                "f3": 64,
                "43": 64,
                "c3": 64,
                "3": 64
            }

        return dict_of_data_length

    @staticmethod
    def hex_to_bin_dict():
        """
        Create a simple dictionary that maps a hex input into a 4 letter binary output.
        :return: dict
        """
        diction = \
            {
                '0': '0000',
                '1': '0001',
                '2': '0010',
                '3': '0011',
                '4': '0100',
                '5': '0101',
                '6': '0110',
                '7': '0111',
                '8': '1000',
                '9': '1001',
                'a': '1010',
                'b': '1011',
                'c': '1100',
                'd': '1101',
                'e': '1110',
                'f': '1111',
            }
        return diction

    def determine_binary(self) -> bool:
        """
        Determine whether we're dealing with a binary or a hex file.
        :return: Boolean value
        """
        if self.filename == '':
            raise ValueError('No filename given.')

        try:
            with open(self.filename, 'r') as f:
                f.read(100)
                is_binary = False
        except UnicodeDecodeError:  # File is binary
            with open(self.filename, 'rb') as f:
                f.read(100)
                is_binary = True
        except FileNotFoundError:
            raise FileNotFoundError("File {} doesn't exist.".format(self.filename))

        else:
            raise ValueError('File [} read unsuccessfully.'.format(self.filename))
        finally:
            return is_binary

    def get_range(self) -> int:
        """
        Finds the "range" of the current file in the proper units
        :return: range as defined my MCS, after bit depth multiplication
        """
        import re

        if self.filename == '':
            raise ValueError('No filename given.')

        if self.is_binary:
            format_str = b'range=(\d+)'
            file_mode = 'rb'
        else:
            format_str = r'range=(\d+)'
            file_mode = 'r'

        format_range = re.compile(format_str)
        with open(self.filename, file_mode) as f:
            cur_str = f.read(500)

        range = int(re.search(format_range, cur_str).group(1))

        return range

    def get_timepatch(self) -> str:
        """
        Get the time patch value out of of a list file.
        :param filename: File to be read.
        :return: Time patch value as string.
        """
        import re

        if self.filename == '':
            raise ValueError('No filename given.')

        if self.is_binary:
            format_str: str = b'time_patch=(\w+)'
            file_mode: str = 'rb'
        else:
            format_str: str = r'time_patch=(\w+)'
            file_mode: str = 'r'

        format_timepatch = re.compile(format_str)
        with open(self.filename, file_mode) as f:
            cur_str: str = f.read(5000)  # read 5000 chars for the timepatch value

        timepatch = re.search(format_timepatch, cur_str).group(1)
        try:
            timepatch = timepatch.decode('utf-8')
        except AttributeError:
            assert isinstance(timepatch, str)
        finally:
            return timepatch

    def find_active_channels(self) -> List[bool]:
        """
        Create a dictionary containing the active channels.
        """
        import re

        if self.filename == '':
            raise ValueError('No filename given.')

        if self.is_binary:
            format_str = b'active=(\d)'
            file_mode = 'rb'
            match = b'1'
        else:
            format_str = r'active=(\d)'
            file_mode = 'r'
            match = '1'

        format_active = re.compile(format_str)
        active_channels: List[bool] = [False, False, False, False, False, False]

        with open(self.filename, file_mode) as f:
            cur_str: str = f.read(5000)

        list_of_matches = re.findall(format_active, cur_str)

        for idx, cur_match in enumerate(list_of_matches):
            if cur_match == match:
                active_channels[idx] = True

        return active_channels

    def get_start_pos(self) -> int:
        """
        Returns the start position of the data
        :param filename: Name of file
        :return: Integer of file position for f.seek() method
        """
        import re

        if self.filename == '':
            raise ValueError('No filename given.')

        if self.is_binary:
            format_str = b'DATA]\r\n'
            file_mode = 'rb'
        else:
            format_str = r'DATA]\n'
            file_mode = 'r'

        format_data = re.compile(format_str)
        pos_in_file: int = 0
        with open(self.filename, file_mode) as f:
            while pos_in_file == 0:
                line = f.readline()
                match = re.search(format_data, line)
                if match is not None:
                    pos_in_file = f.tell()
                    return pos_in_file  # to have the [DATA] as header

    def read_lst(self, num_of_items: int=-1) -> np.ndarray:
        """
        Updated version of LST readout using array slicing (and not Pandas slicing).
        :param num_of_items: Number of lines to read. -1 is all file.
        :return:
        """

        if self.filename is '' or self.start_of_data_pos == 0 or self.timepatch == '':
            return ValueError('Wrong input detected.')

        data_length_dict = self.create_data_length_dict()
        # Make sure we read the exact number of lines we were asked to
        if self.is_binary:
            data_length = data_length_dict[self.timepatch] // 8
        else:
            data_length = data_length_dict[self.timepatch] // 4 + 2

        if num_of_items == -1:
            num_of_lines_to_read = -1
        else:
            num_of_lines_to_read = int(num_of_items * data_length)

        # Read file
        if self.is_binary:
            with open(self.filename, "rb") as f:
                f.seek(self.start_of_data_pos)
                dt: np.dtype = np.dtype(('u1', data_length))
                arr: np.ndarray = np.fromfile(f, dtype=dt, count=num_of_lines_to_read)
                arr_as_bits: np.ndarray = np.unpackbits(np.fliplr(arr), axis=1)
                ########## STILL NOT SUPPORTED ##############
                # It's very hard to pack int32 bits, for example, to a numpy array effectively,
                # since numpy.packbits only works for uint8. Other solutions seem slow and complicated.
                raise NotImplementedError('Binary files are still not supported.')
            return arr_as_bits

        else:
            with open(self.filename, "rb") as f:
                f.seek(self.start_of_data_pos)
                arr = np.fromfile(f, dtype='{}S'.format(data_length),
                                  count=num_of_lines_to_read).astype('{}U'.format(data_length))

            return arr

    def create_inputs_dict(self) -> Dict[str, str]:
        """
        Create a dictionary for all input channels. Currently allows for three channels.
        'Empty' channels will not be checked.
        """
        dict_of_inputs = {}

        if self.input_start != 'Empty':
            dict_of_inputs[self.input_start] = '110'

        if self.input_stop1 != 'Empty':
            dict_of_inputs[self.input_stop1] = '001'

        if self.input_stop2 != 'Empty':
            dict_of_inputs[self.input_stop2] = '010'

        assert len(dict_of_inputs) >= 1
        assert 'Empty' not in list(dict_of_inputs.keys())

        return dict_of_inputs

    def compare_recorded_and_input_channels(self) -> None:
        """
        Raise error if user gave wrong amount of inputs
        """

        # If a user assumes an input exists, but it doesn't - raise an error
        if self.list_of_recorded_data_channels.count(True) < len(self.dict_of_input_channels):
            raise UserWarning('Wrong number of user inputs ({}) compared to number of actual inputs ({}) to the multiscaler.'
                              .format(len(self.dict_of_input_channels), self.list_of_recorded_data_channels.count(True)))

        help_dict = {
            '001': 0,
            '010': 1,
            '110': 2
        }

        for key in self.dict_of_input_channels:
            if not self.list_of_recorded_data_channels[help_dict[self.dict_of_input_channels[key]]]:
                raise UserWarning('Wrong channel specification - the key {} is on an empty channel (number {}).'.\
                      format(key, self.dict_of_input_channels[key]))



# -*- coding: utf-8 -*-
"""
__author__: Hagai Hargil
"""
from collections import OrderedDict
import numpy as np
from typing import Dict, List

class ChoiceManagerHex:

    def __init__(self):
        self.__choice_table = \
        {
            "0": self.timepatch_0,
            "5": self.timepatch_5,
            "1": self.timepatch_1,
            "1a": self.timepatch_1a,
            "2a": self.timepatch_2a,
            "22": self.timepatch_22,
            "32": self.timepatch_32,
            "2": self.timepatch_2,
            "5b": self.timepatch_5b,
            "Db": self.timepatch_Db,
            "f3": self.timepatch_f3,
            "43": self.timepatch_43,
            "c3": self.timepatch_c3,
            "3": self.timepatch_3
        }

    def timepatch_0(self, dict_of_slices):
        dict_of_slices.pop('tag', None)
        dict_of_slices.pop('sweep', None)

        dict_of_slices['abs_time'].start = 0
        dict_of_slices['abs_time'].end = 3
        dict_of_slices['abs_time'].needs_bits = False

        dict_of_slices['chan_edge'].start = 3
        dict_of_slices['chan_edge'].end = 4

        return dict_of_slices

    def timepatch_5(self, dict_of_slices):
        dict_of_slices.pop('tag', None)

        dict_of_slices['sweep'].start = 0
        dict_of_slices['sweep'].end = 2
        dict_of_slices['sweep'].needs_bits = False

        dict_of_slices['abs_time'].start = 2
        dict_of_slices['abs_time'].end = 7
        dict_of_slices['abs_time'].needs_bits = False

        dict_of_slices['chan_edge'].start = 7
        dict_of_slices['chan_edge'].end = 8

        return dict_of_slices

    def timepatch_1(self, dict_of_slices):
        dict_of_slices.pop('tag', None)
        dict_of_slices.pop('sweep', None)

        dict_of_slices['abs_time'].start = 0
        dict_of_slices['abs_time'].end = 7
        dict_of_slices['abs_time'].needs_bits = False

        dict_of_slices['chan_edge'].start = 7
        dict_of_slices['chan_edge'].end = 8

        return dict_of_slices

    def timepatch_1a(self, dict_of_slices):
        dict_of_slices.pop('tag', None)

        dict_of_slices['sweep'].start = 0
        dict_of_slices['sweep'].end = 4
        dict_of_slices['sweep'].needs_bits = False

        dict_of_slices['abs_time'].start = 4
        dict_of_slices['abs_time'].end = 11
        dict_of_slices['abs_time'].needs_bits = False

        dict_of_slices['chan_edge'].start = 11
        dict_of_slices['chan_edge'].end = 12

        return dict_of_slices

    def timepatch_2a(self, dict_of_slices):
        dict_of_slices['tag'].start = 0
        dict_of_slices['tag'].end = 2
        dict_of_slices['tag'].needs_bits = False

        dict_of_slices['sweep'].start = 2
        dict_of_slices['sweep'].end = 4
        dict_of_slices['sweep'].needs_bits = False

        dict_of_slices['abs_time'].start = 4
        dict_of_slices['abs_time'].end = 11
        dict_of_slices['abs_time'].needs_bits = False

        dict_of_slices['chan_edge'].start = 11
        dict_of_slices['chan_edge'].end = 12

        return dict_of_slices

    def timepatch_22(self, dict_of_slices):
        dict_of_slices.pop('sweep', None)

        dict_of_slices['tag'].start = 0
        dict_of_slices['tag'].end = 2
        dict_of_slices['tag'].needs_bits = False

        dict_of_slices['abs_time'].start = 2
        dict_of_slices['abs_time'].end = 11
        dict_of_slices['abs_time'].needs_bits = False

        dict_of_slices['chan_edge'].start = 11
        dict_of_slices['chan_edge'].end = 12

        return dict_of_slices

    def timepatch_32(self, dict_of_slices):
        dict_of_slices['lost'] = True
        dict_of_slices.pop('tag', None)

        dict_of_slices['sweep'].start = 0
        dict_of_slices['sweep'].end = 2
        dict_of_slices['sweep'].needs_bits = True

        dict_of_slices['abs_time'].start = 2
        dict_of_slices['abs_time'].end = 11
        dict_of_slices['abs_time'].needs_bits = False

        dict_of_slices['chan_edge'].start = 11
        dict_of_slices['chan_edge'].end = 12

        return dict_of_slices

    def timepatch_2(self, dict_of_slices):
        dict_of_slices.pop('tag', None)
        dict_of_slices.pop('sweep', None)

        dict_of_slices['abs_time'].start = 0
        dict_of_slices['abs_time'].end = 11
        dict_of_slices['abs_time'].needs_bits = False

        dict_of_slices['chan_edge'].start = 11
        dict_of_slices['chan_edge'].end = 12

        return dict_of_slices

    def timepatch_5b(self, dict_of_slices):
        dict_of_slices['lost'] = True

        dict_of_slices['tag'].start = 0
        dict_of_slices['tag'].end = 4
        dict_of_slices['tag'].needs_bits = True

        dict_of_slices['sweep'].start = 4
        dict_of_slices['sweep'].end = 8
        dict_of_slices['sweep'].needs_bits = False

        dict_of_slices['abs_time'].start = 8
        dict_of_slices['abs_time'].end = 15
        dict_of_slices['abs_time'].needs_bits = False

        dict_of_slices['chan_edge'].start = 15
        dict_of_slices['chan_edge'].end = 16

        return dict_of_slices

    def timepatch_Db(self, dict_of_slices):
        dict_of_slices['tag'].start = 0
        dict_of_slices['tag'].end = 4
        dict_of_slices['tag'].needs_bits = False

        dict_of_slices['sweep'].start = 4
        dict_of_slices['sweep'].end = 8
        dict_of_slices['sweep'].needs_bits = False

        dict_of_slices['abs_time'].start = 8
        dict_of_slices['abs_time'].end = 15
        dict_of_slices['abs_time'].needs_bits = False

        dict_of_slices['chan_edge'].start = 15
        dict_of_slices['chan_edge'].end = 16

        return dict_of_slices

    def timepatch_f3(self, dict_of_slices):
        dict_of_slices['lost'] = True

        dict_of_slices['tag'].start = 0
        dict_of_slices['tag'].end = 4
        dict_of_slices['tag'].needs_bits = False

        dict_of_slices['sweep'].start = 4
        dict_of_slices['sweep'].end = 6
        dict_of_slices['sweep'].needs_bits = True

        dict_of_slices['abs_time'].start = 6
        dict_of_slices['abs_time'].end = 15
        dict_of_slices['abs_time'].needs_bits = False

        dict_of_slices['chan_edge'].start = 15
        dict_of_slices['chan_edge'].end = 16

        return dict_of_slices

    def timepatch_43(self, dict_of_slices):
        dict_of_slices.pop('sweep', None)

        dict_of_slices['lost'] = True

        dict_of_slices['tag'].start = 0
        dict_of_slices['tag'].end = 4
        dict_of_slices['tag'].needs_bits = True

        dict_of_slices['abs_time'].start = 4
        dict_of_slices['abs_time'].end = 15
        dict_of_slices['abs_time'].needs_bits = False

        dict_of_slices['chan_edge'].start = 15
        dict_of_slices['chan_edge'].end = 16

        return dict_of_slices

    def timepatch_c3(self, dict_of_slices):
        dict_of_slices.pop('sweep', None)

        dict_of_slices['tag'].start = 0
        dict_of_slices['tag'].end = 4
        dict_of_slices['tag'].needs_bits = False

        dict_of_slices['abs_time'].start = 4
        dict_of_slices['abs_time'].end = 15
        dict_of_slices['abs_time'].needs_bits = False

        dict_of_slices['chan_edge'].start = 15
        dict_of_slices['chan_edge'].end = 16

        return dict_of_slices

    def timepatch_3(self, dict_of_slices):
        # NOT IMPLEMENTED IN THE ACTUAL CODE
        dict_of_slices.pop('sweep', None)

        dict_of_slices['tag'].start = 0
        dict_of_slices['tag'].end = 2
        dict_of_slices['tag'].needs_bits = True

        dict_of_slices['abs_time'].start = 2
        dict_of_slices['abs_time'].end = 15
        dict_of_slices['abs_time'].needs_bits = False

        dict_of_slices['chan_edge'].start = 15
        dict_of_slices['chan_edge'].end = 16

        return dict_of_slices

    def __generate_dict_from_indices(self) -> OrderedDict:
        """
        Generate a basic template dictionary to be populated with slicing indices.
        :param slices:
        :return:
        """
        from pysight.movie_tools import Struct
        slice_dict: OrderedDict = OrderedDict()
        keys = ['lost', 'tag', 'sweep', 'abs_time', 'chan_edge']

        slice_dict[keys[0]] = False
        for key in keys[1:]:
            slice_dict[key] = Struct(np.nan, np.nan)

        return slice_dict

    def process(self, case):
        """
        Generate a dictionary and populate it with the proper value, accoring to the timepatch (==case).
        """
        assert isinstance(case, str)

        dict_of_slices = self.__generate_dict_from_indices()
        populated_dict_of_slices = self.__choice_table[case](dict_of_slices)

        return populated_dict_of_slices


class ChoiceManagerBinary:

    def __init__(self) -> None:
        self.__choice_table = \
        {
            "0": self.timepatch_0,
            "5": self.timepatch_5,
            "1": self.timepatch_1,
            "1a": self.timepatch_1a,
            "2a": self.timepatch_2a,
            "22": self.timepatch_22,
            "32": self.timepatch_32,
            "2": self.timepatch_2,
            "5b": self.timepatch_5b,
            "Db": self.timepatch_Db,
            "f3": self.timepatch_f3,
            "43": self.timepatch_43,
            "c3": self.timepatch_c3,
            "3": self.timepatch_3
        }

    def timepatch_0(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop('tag', None)
        dict_of_slices.pop('sweep', None)

        dict_of_slices['abs_time'].start = 0
        dict_of_slices['abs_time'].end = 12
        dict_of_slices['abs_time'].cols = 16

        dict_of_slices['edge'].start = 12
        dict_of_slices['edge'].end = 13
        dict_of_slices['edge'].cols = 8

        dict_of_slices['chan'].start = 13
        dict_of_slices['chan'].end = 15
        dict_of_slices['chan'].cols = 8

        return dict_of_slices

    def timepatch_5(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop('tag', None)

        dict_of_slices['sweep'].start = 0
        dict_of_slices['sweep'].end = 8

        dict_of_slices['abs_time'].start = 8
        dict_of_slices['abs_time'].end = 28

        dict_of_slices['chan_edge'].start = 28
        dict_of_slices['chan_edge'].end = 32

        return dict_of_slices

    def timepatch_1(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop('tag', None)
        dict_of_slices.pop('sweep', None)

        dict_of_slices['abs_time'].start = 0
        dict_of_slices['abs_time'].end = 28

        dict_of_slices['chan_edge'].start = 28
        dict_of_slices['chan_edge'].end = 32

        return dict_of_slices

    def timepatch_1a(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop('tag', None)

        dict_of_slices['sweep'].start = 0
        dict_of_slices['sweep'].end = 16

        dict_of_slices['abs_time'].start = 16
        dict_of_slices['abs_time'].end = 44

        dict_of_slices['chan_edge'].start = 44
        dict_of_slices['chan_edge'].end = 48

        return dict_of_slices

    def timepatch_2a(self, dict_of_slices) -> OrderedDict:
        dict_of_slices['tag'].start = 0
        dict_of_slices['tag'].end = 8

        dict_of_slices['sweep'].start = 8
        dict_of_slices['sweep'].end = 16

        dict_of_slices['abs_time'].start = 16
        dict_of_slices['abs_time'].end = 44

        dict_of_slices['chan_edge'].start = 44
        dict_of_slices['chan_edge'].end = 48

        return dict_of_slices

    def timepatch_22(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop('sweep', None)

        dict_of_slices['tag'].start = 0
        dict_of_slices['tag'].end = 8

        dict_of_slices['abs_time'].start = 8
        dict_of_slices['abs_time'].end = 44

        dict_of_slices['chan_edge'].start = 44
        dict_of_slices['chan_edge'].end = 48

        return dict_of_slices

    def timepatch_32(self, dict_of_slices) -> OrderedDict:
        dict_of_slices['lost'] = 1
        dict_of_slices.pop('tag', None)

        dict_of_slices['sweep'].start = 1
        dict_of_slices['sweep'].end = 8

        dict_of_slices['abs_time'].start = 8
        dict_of_slices['abs_time'].end = 44

        dict_of_slices['chan_edge'].start = 44
        dict_of_slices['chan_edge'].end = 48

        return dict_of_slices

    def timepatch_2(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop('tag', None)
        dict_of_slices.pop('sweep', None)

        dict_of_slices['abs_time'].start = 0
        dict_of_slices['abs_time'].end = 44

        dict_of_slices['chan_edge'].start = 44
        dict_of_slices['chan_edge'].end = 48

        return dict_of_slices

    def timepatch_5b(self, dict_of_slices) -> OrderedDict:
        dict_of_slices['lost'].start = 0
        dict_of_slices['lost'].end = 1
        dict_of_slices['lost'].cols = 8

        dict_of_slices['tag'].start = 1
        dict_of_slices['tag'].end = 16
        dict_of_slices['tag'].cols = 16

        dict_of_slices['sweep'].start = 16
        dict_of_slices['sweep'].end = 32

        dict_of_slices['abs_time'].start = 32
        dict_of_slices['abs_time'].end = 60
        dict_of_slices['abs_time'].cols = 32

        dict_of_slices['edge'].start = 60
        dict_of_slices['edge'].end = 61
        dict_of_slices['edge'].cols = 8

        dict_of_slices['chan'].start = 61
        dict_of_slices['chan'].end = 64
        dict_of_slices['chan'].cols = 8

        return dict_of_slices

    def timepatch_Db(self, dict_of_slices) -> OrderedDict:
        dict_of_slices['tag'].start = 0
        dict_of_slices['tag'].end = 16

        dict_of_slices['sweep'].start = 16
        dict_of_slices['sweep'].end = 32

        dict_of_slices['abs_time'].start = 32
        dict_of_slices['abs_time'].end = 60

        dict_of_slices['chan_edge'].start = 60
        dict_of_slices['chan_edge'].end = 64

        return dict_of_slices

    def timepatch_f3(self, dict_of_slices) -> OrderedDict:
        dict_of_slices['lost'].start = 0
        dict_of_slices['lost'].end = 1
        dict_of_slices['lost'].cols = 8

        dict_of_slices['tag'].start = 1
        dict_of_slices['tag'].end = 17
        dict_of_slices['tag'].cols = 16

        dict_of_slices['sweep'].start = 17
        dict_of_slices['sweep'].end = 24
        dict_of_slices['sweep'].cols = 8

        dict_of_slices['abs_time'].start = 24
        dict_of_slices['abs_time'].end = 60
        dict_of_slices['abs_time'].cols = 40

        dict_of_slices['edge'].start = 60
        dict_of_slices['edge'].end = 61
        dict_of_slices['edge'].cols = 8

        dict_of_slices['chan'].start = 61
        dict_of_slices['chan'].end = 64
        dict_of_slices['chan'].cols = 8

        return dict_of_slices

    def timepatch_43(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop('sweep', None)

        dict_of_slices['lost'] = 1

        dict_of_slices['tag'].start = 1
        dict_of_slices['tag'].end = 16

        dict_of_slices['abs_time'].start = 16
        dict_of_slices['abs_time'].end = 60

        dict_of_slices['chan_edge'].start = 60
        dict_of_slices['chan_edge'].end = 64

        return dict_of_slices

    def timepatch_c3(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop('sweep', None)

        dict_of_slices['tag'].start = 0
        dict_of_slices['tag'].end = 16

        dict_of_slices['abs_time'].start = 16
        dict_of_slices['abs_time'].end = 60

        dict_of_slices['chan_edge'].start = 60
        dict_of_slices['chan_edge'].end = 64

        return dict_of_slices

    def timepatch_3(self, dict_of_slices) -> OrderedDict:
        dict_of_slices.pop('sweep', None)

        dict_of_slices['lost'] = 1

        dict_of_slices['tag'].start = 1
        dict_of_slices['tag'].end = 6

        dict_of_slices['abs_time'].start = 6
        dict_of_slices['abs_time'].end = 60

        dict_of_slices['chan_edge'].start = 60
        dict_of_slices['chan_edge'].end = 64

        return dict_of_slices

    def __generate_dict_from_indices(self) -> OrderedDict:
        """
        Generate a basic template dictionary to be populated with slicing indices.
        :param slices:
        :return:
        """
        from pysight.movie_tools import Struct
        slice_dict: OrderedDict = OrderedDict()
        keys: List[str] = ['lost', 'tag', 'sweep', 'abs_time', 'edge', 'chan']

        for key in keys:
            slice_dict[key] = Struct(np.nan, np.nan)

        return slice_dict

    def process(self, case: str) -> OrderedDict:
        """
        Generate a dictionary and populate it with the proper value, accoring to the timepatch (==case).
        """
        assert isinstance(case, str)

        dict_of_slices: OrderedDict = self.__generate_dict_from_indices()
        populated_dict_of_slices: OrderedDict = self.__choice_table[case](dict_of_slices)

        return populated_dict_of_slices




"""
__author__ = Hagai Hargil
"""


import unittest
from pysight.censor_tools import CensorCorrection
from pysight.censor_tools import CensoredVolume
import pandas as pd
from collections import deque, namedtuple
import numpy as np


class TestCensorTools(unittest.TestCase):
    length = 100
    df = pd.DataFrame([i for i in range(length)], columns=['abs_time'])
    tuptype = namedtuple('TestCensor', ('hist', 'edges'))
    cens = tuptype
    cens.hist = np.array([i for i in range(length)])
    cens.edges = np.array([1, 4, 8, 18, 25, length - 1])

    def test_min_time(self):
        censored = CensorCorrection(df=self.df, reprate=80e6,
                                    deque_of_vols=deque([self.cens]),
                                    binwidth=800e-12, offset=9)
        self.assertEqual(censored.start_time, 0)

    def test_max_time(self):
        censored = CensorCorrection(df=self.df, reprate=80e6,
                                    deque_of_vols=deque([self.cens]),
                                    binwidth=800e-12, offset=9)
        self.assertEqual(censored.end_time, self.length - 1 + 16)

    def test_laser_pulses(self):
        censored = CensorCorrection(df=self.df, reprate=80e6,
                                    deque_of_vols=deque([self.cens]),
                                    binwidth=800e-12, offset=9)
        real_range = np.arange(start=0+censored.offset, stop=self.length-1+16, step=16, dtype=np.uint64)
        self.assertListEqual(list(real_range), list(censored.laser_pulses))

    def test_bincount(self):
        censored = CensorCorrection(df=self.df, reprate=80e6,
                                    deque_of_vols=deque([self.cens]),
                                    binwidth=800e-12, offset=9)
        real_result = np.bincount(np.histogram(self.cens.hist, bins=np.arange(start=0+censored.offset,
                                                                              stop=self.length-1+16,
                                                                              step=16, dtype=np.uint64))[0])
        self.assertListEqual(list(real_result), list(censored.gen_bincount()))

    def test_sort_photons(self):
        # censored = CensorCorrection(df=self.df, reprate=80e6,
        #                             deque_of_vols=deque([self.cens]),
        #                             binwidth=800e-12, offset=9)
        #
        # real_pulses = np.arange(start=0+censored.offset, stop=99+16, step=16, dtype=np.uint64)
        # sorted_indices = np.searchsorted(np.arange(start=9, stop=99+16, step=16, dtype=np.uint64), [i for i in range(100)])
        pass
